var pg = require('pg').native;
var _ = require('underscore');

//-------------------------------

var Query = function(){
	this._fields = {};
	this._joins = {};
	this._filters = [];
};

Query.prototype.table = function(tableName, tableAlias){
	this._tableName = tableName;
	this._tableAlias = tableAlias ? tableAlias : '_self';
	return this;
};

Query.prototype.field = function(fieldName, sqlExpr, params){
	if(sqlExpr instanceof Array){
		params = sqlExpr;
		sqlExpr = '$?';
	}
	this._fields[fieldName] = { sqlExpr: sqlExpr, params: params };
	return this;
};

Query.prototype.leftJoin = function(joinName, tableName, sqlCond, params){
	this._joins[joinName] = {
		tableName: tableName,
		sqlCond: sqlCond,
		params: params,
		type: 'left outer'
	};
	return this;
};

Query.prototype.rightJoin = function(joinName, tableName, sqlCond, params){
	this._joins[joinName] = {
		tableName: tableName,
		sqlCond: sqlCond,
		params: params,
		type: 'right outer'
	};
	return this;
};

Query.prototype.fullJoin = function(joinName, tableName, sqlCond, params){
	this._joins[joinName] = {
		tableName: tableName,
		sqlCond: sqlCond,
		params: params,
		type: 'full outer'
	};
	return this;
};

Query.prototype.innerJoin = function(joinName, tableName, sqlCond, params){
	this._joins[joinName] = {
		tableName: tableName,
		sqlCond: sqlCond,
		params: params,
		type: 'inner'
	};
	return this;
};

Query.prototype.crossJoin = function(joinName, tableName){
	this._joins[joinName] = {
		tableName: tableName,
		type: 'cross'
	};
	return this;
};

Query.prototype.filter = function(sqlExpr, params){
	this._filters.push({
		sqlExpr: sqlExpr,
		params: params
	});
	return this;
};

Query.prototype.limit = function(limit){
	this._limit = limit;
	return this;
};

Query.prototype.offset = function(offset){
	this._offset = offset;
	return this;
};

Query.prototype.order = function(order){
	this._order = order;
	return this;
};

Query.prototype._prepareSelectSql = function(mode){
	var fieldsSql = '';
	for(var fieldName in this._fields){
		var sqlExpr = this._fields[fieldName].sqlExpr;
		if(fieldsSql.length)
			fieldsSql += ',\n';
		fieldsSql += sqlExpr + ' as "' + fieldName + '"';
	}

	var params = [];
	var paramsCount = 0;
	
	var joinsSql = '';
	for(var joinName in this._joins){
		var join = this._joins[joinName];
		if(join.params)
			params = params.concat(join.params);
		var sqlCond = join.sqlCond.replace(/\$\?/g, function(str){
			return '$' + (++ paramsCount);
		});
		joinsSql += '\n' + join.type + ' join "' + join.tableName + '" as "' + joinName + '" on ' + sqlCond;
	}
	
	var filtersSql = '';
	this._filters.forEach(function(filter){
		if(filter.params)
			params = params.concat(filter.params);
		var sqlExpr = filter.sqlExpr.replace(/\$\?/g, function(str){
			return '$' + (++ paramsCount);
		});
		filtersSql += (filtersSql.length ? ' and ' : '') + '(' + sqlExpr + ')';
	});
	if(filtersSql.length)
		filtersSql = '\nwhere ' +filtersSql;
	
	var sql = 'select\n' + fieldsSql
		+ '\nfrom "' + this._tableName + '" as "' + this._tableAlias + '"'
		+ joinsSql
		+ filtersSql;
	
	if(this._order && mode != 'count')
		sql += '\norder by ' + this._order;
	
	if(this._offset)
		sql += '\noffset ' + this._offset;
	
	if(mode == 'row')
		sql += '\nlimit 1';
	else if(this._limit)
		sql += '\nlimit ' + this._limit;

	return {
		sql: sql,
		params: params
	};
};

Query.prototype.rows = function(dbConnect, next){
	var prepared = this._prepareSelectSql();
	dbConnect.query(prepared.sql, prepared.params, function(err, result){
		next(err, err ? [] : result.rows);
	});
	return this;
};

Query.prototype.row = function(dbConnect, next){
	var prepared = this._prepareSelectSql('row');
	dbConnect.query(prepared.sql, prepared.params, function(err, result){
		if(err)
			next(err);
		else if(!result.rows.length)
			next(new Error('Record not found.'));
		else
			next(null, result.rows[0]);
	});
	return this;
};

Query.prototype.count = function(dbConnect, next){
	var prepared = this._prepareSelectSql('count');
	dbConnect.query('select count(*) as _count from (' + prepared.sql + ') as tbl', prepared.params, function(err, result){
		next(err, err ? 0 : result.rows[0]['_count']);
	});
	return this;
};

Query.prototype.update = function(dbConnect, next){
	var prepared = this._prepareUpdateSql();
	dbConnect.query(prepared.sql, prepared.params, next);
	return this;
};

Query.prototype.del = function(dbConnect, next){
	var prepared = this._prepareDeleteSql();
	dbConnect.query(prepared.sql, prepared.params, next);
	return this;
};

Query.prototype.clone = function(){
	var q = Object.create(Query.prototype);
	for(keyName in this)
		if(this.hasOwnProperty(keyName))
			if(this[keyName] instanceof Object)
				q[keyName] = _.clone(this[keyName]);
			else
				q[keyName] = this[keyName];
	return q;
	//return _.clone(this);
	//return _.extend(Object.create(Query.prototype), this);
};

Query.prototype._prepareUpdateSql = function(){
	var params = [];
	var paramsCount = 0;

	var fieldsSql = '';
	for(var fieldName in this._fields){
		var field = this._fields[fieldName];
		if(field.params)
			params = params.concat(field.params);
		var sqlExpr = field.sqlExpr.replace(/\$\?/g, function(str){
			return '$' + (++ paramsCount);
		});
		if(fieldsSql.length)
			fieldsSql += ',\n';
		fieldsSql += '"' + fieldName + '" = ' + sqlExpr;
	}
	
	var firstInnerJoinName = false;
	var joinsSql = '';
	for(var joinName in this._joins){
		var join = this._joins[joinName];
		if(!firstInnerJoinName && join.type == 'inner'){
			firstInnerJoinName = joinName;
			continue;
		}
		if(join.params)
			params = params.concat(join.params);
		var sqlCond = join.sqlCond.replace(/\$\?/g, function(str){
			return '$' + (++ paramsCount);
		});
		joinsSql += '\n' + join.type + ' join "' + join.tableName + '" as "' + joinName + '" on ' + sqlCond;
	}
	
	var filtersSql = '';
	if(firstInnerJoinName){
		var firstInnerJoin = this._joins[firstInnerJoinName];
		if(firstInnerJoin.params)
			params = params.concat(firstInnerJoin.params);
		var sqlExpr = firstInnerJoin.sqlCond.replace(/\$\?/g, function(str){
			return '$' + (++ paramsCount);
		});
		filtersSql += '(' + sqlExpr + ')';
	}
	this._filters.forEach(function(filter){
		if(filter.params)
			params = params.concat(filter.params);
		var sqlExpr = filter.sqlExpr.replace(/\$\?/g, function(str){
			return '$' + (++ paramsCount);
		});
		filtersSql += (filtersSql.length ? ' and ' : '') + '(' + sqlExpr + ')';
	});
	if(filtersSql.length)
		filtersSql = '\nwhere ' + filtersSql;

	var firstInnerJoinSql = '';
	if(firstInnerJoinName)
		firstInnerJoinSql = '\nfrom "' + this._joins[firstInnerJoinName].tableName + '" as "' + firstInnerJoinName + '"';
	
	var sql = 'update "' + this._tableName + '" as "' + this._tableAlias + '" set\n'
	+ fieldsSql
	+ firstInnerJoinSql
	+ joinsSql
	+ filtersSql;

	return {
		sql: sql,
		params: params
	};
};

Query.prototype._prepareDeleteSql = function(){
	var params = [];
	var paramsCount = 0;

	var filtersSql = '';
	var usingSql = '';
	for(var joinName in this._joins){
		var join = this._joins[joinName];
		if(join.params)
			params = params.concat(join.params);
		var sqlCond = join.sqlCond.replace(/\$\?/g, function(str){
			return '$' + (++ paramsCount);
		});
		filtersSql += (filtersSql.length ? ' and ' : '') + '(' + sqlCond + ')';
		usingSql += (usingSql.length ? ',\n' : '') + '"' + join.tableName + '" as "' + joinName;
	}
	if(usingSql.length)
		usingSql = '\nusing\n' + usingSql;
	
	this._filters.forEach(function(filter){
		if(filter.params)
			params = params.concat(filter.params);
		var sqlExpr = filter.sqlExpr.replace(/\$\?/g, function(str){
			return '$' + (++ paramsCount);
		});
		filtersSql += (filtersSql.length ? ' and ' : '') + '(' + sqlExpr + ')';
	});
	if(filtersSql.length)
		filtersSql = '\nwhere ' + filtersSql;
	
	var sql = 'delete from "' + this._tableName + '" as "' + this._tableAlias + '"'
	+ usingSql
	+ filtersSql;

	return {
		sql: sql,
		params: params
	};
};

//-------------------------------

var allTables = {};

function Table(params){
	if(typeof params == 'string')
		return allTables[params];

	this._params = params;
	allTables[params.tableName] = this;
	
	_.defaults(this._params, {
		primaryKey: 'id',
		fields: {},
		relations: {},
	});

}

Table.prototype.field = function(name, field){
	this._params.fields[name] = field;
};

Table.prototype.relation = function(name, relation){
	this._params.relations[name] = relation;
};

Table.prototype.slice = function(paths){
	var slice = {
		fields: {},
		query: (new Query).table(this._params.tableName)
	};

	var self = this;
	paths.forEach(function(path){
		var pathParts = path.split('.');
		var fieldName = pathParts.pop();

		var prefix = [];
		var table = self;
		pathParts.forEach(function(part){
			var rel = table._params.relations[part];
			prefix.push(part);
			table = new Table(rel.toTable);
		});
		if(fieldName == '*')
			for(fieldName in table._params.fields)
				slice.fields[prefix.concat([fieldName]).join('.')] = table._params.fields[fieldName];
		else
			slice.fields[prefix.concat([fieldName]).join('.')] = table._params.fields[fieldName];
	});
	
	for(path in slice.fields){
		var pathParts = path.split('.');
		var fieldName = pathParts.pop();

		var prefix = [];
		var table = this;
		pathParts.forEach(function(part){
			var rel = table._params.relations[part];
			table = new Table(rel.toTable);
			var tabAlias = ['_self'].concat(prefix).join('.');
			prefix.push(part);
			var relAlias = ['_self'].concat(prefix).join('.');
			slice.query.leftJoin(relAlias, rel.toTable, '"'+relAlias+'".'+table._params.primaryKey+' = "'+tabAlias+'".'+rel.fromField);
		});
		slice.query.field(path, '"'+['_self'].concat(prefix).join('.')+'".'+fieldName);
	}

	
	return slice;
};

//-------------------------------

module.exports.Query = Query;
module.exports.Table = Table;
module.exports.connect = pg.connect;

